/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Alex Leigh <php (at) postfin (dot) com>                      |
  +----------------------------------------------------------------------+
*/

/*
  TODO:
     suppress errors on @
     detect hangup and abort.
     don't crash on php critical errors. (like a missing function)
     test all functions with bad input
*/


/*
 * ZTS needs to be defined for Asterisk to work
 */
#if !defined(ZTS)
#error "Asterisk sapi module needs ZTS to be defined"
#endif

#ifndef _REENTRANT
#define _REENTRANT
#endif

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif


/* PHP includes */
#include "php.h"
#include "php_variables.h"
#include "ext/standard/info.h"
#include "php_ini.h"
#include "php_globals.h"
#include "SAPI.h"
#include "php_main.h"
#include "php_version.h"
#include "TSRM.h"
#include "ext/standard/php_standard.h"

/* Asterisk includes */
#include <asterisk/pbx.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/module.h>
#include <asterisk/options.h>
#include <asterisk/utils.h>
#include <asterisk/features.h>
#include <asterisk/callerid.h>

STANDARD_LOCAL_USER;
LOCAL_USER_DECL;

/* *********************************************************************** *
 * *            PHP Module                                               * *
 * *********************************************************************** */

#define MAX_ARGS 128

/*
 * Structure to encapsulate the asterisk request in SAPI
 */
typedef struct php_asterisk_globals {
	struct ast_channel *channel;
	int argc;
	char *argv[MAX_ARGS+1];
} php_asterisk_globals;
#define ASTG(v)		(request_context->v)

#ifdef POST_10
#define ast_set_read_format_a(chan,fmt) ast_set_read_format(chan,fmt,0);
#define ast_set_write_format_a(chan,fmt) ast_set_write_format(chan,fmt,0);
#else
#define ast_set_read_format_a(chan,fmt) ast_set_read_format(chan,fmt);
#define ast_set_write_format_a(chan,fmt) ast_set_write_format(chan,fmt);
#endif

PHP_MINIT_FUNCTION(asterisk);
PHP_MINFO_FUNCTION(asterisk);

PHP_FUNCTION(asterisk_answer);
PHP_FUNCTION(asterisk_best_codec);
PHP_FUNCTION(asterisk_bridge_call);
PHP_FUNCTION(asterisk_chanlist);
PHP_FUNCTION(asterisk_channel_make_compatible);
PHP_FUNCTION(asterisk_channel_setoption);
PHP_FUNCTION(asterisk_channel_status);
PHP_FUNCTION(asterisk_cli);
PHP_FUNCTION(asterisk_codec2str);
PHP_FUNCTION(asterisk_control_streamfile);
PHP_FUNCTION(asterisk_db_del);
PHP_FUNCTION(asterisk_db_deltree);
PHP_FUNCTION(asterisk_db_get);
PHP_FUNCTION(asterisk_db_put);
PHP_FUNCTION(asterisk_exec);
PHP_FUNCTION(asterisk_getdata);
PHP_FUNCTION(asterisk_getformatbyname);
PHP_FUNCTION(asterisk_getformatname);
PHP_FUNCTION(asterisk_get_priority);
PHP_FUNCTION(asterisk_get_variable);
PHP_FUNCTION(asterisk_get_variables);
PHP_FUNCTION(asterisk_hangup);
PHP_FUNCTION(asterisk_log);
PHP_FUNCTION(asterisk_manager_command);
PHP_FUNCTION(asterisk_moh_start);
PHP_FUNCTION(asterisk_moh_stop);
PHP_FUNCTION(asterisk_recvchar);
PHP_FUNCTION(asterisk_request);
PHP_FUNCTION(asterisk_request_and_dial);
PHP_FUNCTION(asterisk_say_character_str);
PHP_FUNCTION(asterisk_say_date);
PHP_FUNCTION(asterisk_say_datetime);
PHP_FUNCTION(asterisk_say_datetime_from_now);
PHP_FUNCTION(asterisk_say_digit_str);
PHP_FUNCTION(asterisk_say_number);
PHP_FUNCTION(asterisk_say_phonetic_str);
PHP_FUNCTION(asterisk_say_time);
PHP_FUNCTION(asterisk_senddigit);
PHP_FUNCTION(asterisk_send_image);
PHP_FUNCTION(asterisk_sendtext);
PHP_FUNCTION(asterisk_set_best_read_format);
PHP_FUNCTION(asterisk_set_best_write_format);
PHP_FUNCTION(asterisk_set_callerid);
PHP_FUNCTION(asterisk_set_context);
PHP_FUNCTION(asterisk_set_extension);
PHP_FUNCTION(asterisk_set_priority);
PHP_FUNCTION(asterisk_set_read_format);
PHP_FUNCTION(asterisk_set_variable);
PHP_FUNCTION(asterisk_setwhentohangup);
PHP_FUNCTION(asterisk_set_write_format);
PHP_FUNCTION(asterisk_stream_file);
PHP_FUNCTION(asterisk_state2str);
PHP_FUNCTION(asterisk_tdd_mode);
PHP_FUNCTION(asterisk_verbose);
PHP_FUNCTION(asterisk_waitfor);
PHP_FUNCTION(asterisk_waitfordigit);

/* {{{ asterisk_functions[]
 *
 * Every user visible function must have an entry in asterisk_functions[].
 */
function_entry asterisk_functions[] = {
	PHP_FE(asterisk_answer,				NULL)
	PHP_FE(asterisk_best_codec,			NULL)
	PHP_FE(asterisk_bridge_call,			NULL)
	PHP_FE(asterisk_chanlist,			NULL)
	PHP_FE(asterisk_channel_make_compatible,	NULL)
	PHP_FE(asterisk_channel_setoption,		NULL)
	PHP_FE(asterisk_channel_status,			NULL)
	PHP_FE(asterisk_cli,				NULL)
	PHP_FE(asterisk_codec2str,			NULL)
	PHP_FE(asterisk_control_streamfile,		NULL)
	PHP_FE(asterisk_db_del,				NULL)
	PHP_FE(asterisk_db_deltree,			NULL)
	PHP_FE(asterisk_db_get,				NULL)
	PHP_FE(asterisk_db_put,				NULL)
	PHP_FE(asterisk_exec,				NULL)
	PHP_FE(asterisk_getdata,			NULL)
	PHP_FE(asterisk_getformatbyname,		NULL)
	PHP_FE(asterisk_getformatname,			NULL)
	PHP_FE(asterisk_get_priority,			NULL)
	PHP_FE(asterisk_get_variable,			NULL)
	PHP_FE(asterisk_get_variables,			NULL)
	PHP_FE(asterisk_hangup,				NULL)
	PHP_FE(asterisk_log,				NULL)
	PHP_FE(asterisk_manager_command,		NULL)
	PHP_FE(asterisk_moh_start,			NULL)
	PHP_FE(asterisk_moh_stop,			NULL)
	PHP_FE(asterisk_recvchar,			NULL)
	PHP_FE(asterisk_request,			NULL)
	PHP_FE(asterisk_request_and_dial,		NULL)
	PHP_FE(asterisk_say_character_str,		NULL)
	PHP_FALIAS(asterisk_say_character, asterisk_say_digit_str, NULL)
	PHP_FE(asterisk_say_date,			NULL)
	PHP_FE(asterisk_say_datetime,			NULL)
	PHP_FE(asterisk_say_datetime_from_now,		NULL)
	PHP_FE(asterisk_say_digit_str,			NULL)
	PHP_FALIAS(asterisk_say_digit, asterisk_say_digit_str, NULL)
	PHP_FE(asterisk_say_number,			NULL)
	PHP_FE(asterisk_say_phonetic_str,		NULL)
	PHP_FALIAS(asterisk_say_phonetic, asterisk_say_phonetic_str, NULL)
	PHP_FE(asterisk_say_time,			NULL)
	PHP_FE(asterisk_senddigit,			NULL)
	PHP_FE(asterisk_send_image,			NULL)
	PHP_FE(asterisk_sendtext,			NULL)
	PHP_FE(asterisk_set_best_read_format,		NULL)
	PHP_FE(asterisk_set_best_write_format,		NULL)
	PHP_FE(asterisk_set_callerid,			NULL)
	PHP_FE(asterisk_set_context,			NULL)
	PHP_FE(asterisk_set_extension,			NULL)
	PHP_FE(asterisk_set_priority,			NULL)
	PHP_FE(asterisk_set_read_format,		NULL)
	PHP_FE(asterisk_set_variable,			NULL)
	PHP_FE(asterisk_setwhentohangup,		NULL)
	PHP_FALIAS(asterisk_set_autohangup, asterisk_setwhentohangup, NULL)
	PHP_FE(asterisk_set_write_format,		NULL)
	PHP_FE(asterisk_stream_file,			NULL)
	PHP_FE(asterisk_state2str,			NULL)
	PHP_FE(asterisk_tdd_mode,			NULL)
	PHP_FE(asterisk_verbose,			NULL)
	PHP_FE(asterisk_waitfor,			NULL)
	PHP_FE(asterisk_waitfordigit,			NULL)
	{ NULL, NULL, NULL }
};
/* }}} */


/* {{{ asterisk_module_entry
 */
zend_module_entry asterisk_module_entry = {
	STANDARD_MODULE_HEADER,
	"asterisk",
	asterisk_functions,
	PHP_MINIT(asterisk),
	NULL,
	NULL,
	NULL,
	PHP_MINFO(asterisk),
	NO_VERSION_YET,
	STANDARD_MODULE_PROPERTIES
};
/* }}} */


/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINIT_FUNCTION(asterisk)
{
/*
	REGISTER_STRING_CONSTANT("JELLO", "food", CONST_CS);
*/

        return SUCCESS;
}
/* }}} */



/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(asterisk)
{
	php_info_print_box_start(0);
	php_printf("<p>Asterisk PHP extension</p>\n");
	php_printf("<p>Version 0.1alpha (2005-06-09)</p>\n");
	php_printf("<p><b>Authors:</b></p>\n");
	php_printf("<p>David Eder &lt;david@eder.us&gt;</p>\n");
	php_info_print_box_end();

	DISPLAY_INI_ENTRIES();
}
/* }}} */


struct ast_channel *asterisk_get_channel_by_name(char *name)
{
	struct ast_channel *chan;

	if(name) {
		chan = ast_get_channel_by_name_locked(name);
		if(chan) {
			ast_mutex_unlock(&chan->lock);
		}
	} else {
		TSRMLS_FETCH();
		chan = ((php_asterisk_globals *) SG(server_context))->channel;
	}
	return(chan);
}

/* {{{ proto int asterisk_answer([string channel])
  Answer a channel */
PHP_FUNCTION(asterisk_answer)
{
	zval **args[1];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 1) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(chan->_state != AST_STATE_UP) {
		RETURN_LONG(ast_answer(chan));
	}

	RETURN_LONG(AST_STATE_UP);
}
/* }}} */


/* {{{ proto int asterisk_best_codec([string channel])
  Get best compatible code for this channel */
PHP_FUNCTION(asterisk_best_codec)
{
	zval **args[1];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 1) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	RETURN_LONG(ast_best_codec(chan->nativeformats));
}
/* }}} */


/* {{{ proto int asterisk_bridge_call([string channel,] string peer[, array options])
  Bridge a call */
PHP_FUNCTION(asterisk_bridge_call)
{
/* TODO: test */
	zval **args[3];
	zval *options;
	zval **zvalue;
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;
	struct ast_channel * pr;

	struct ast_bridge_config config;

	if(arg_count > 3 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	convert_to_string_ex(args[0]);
	if(arg_count < 3) {
		chan = asterisk_get_channel_by_name(NULL);
		pr = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	} else {
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_string_ex(args[1]);
		pr = asterisk_get_channel_by_name(Z_STRVAL_PP(args[1]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(!pr) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	memset(&config, 0, sizeof(struct ast_bridge_config));

	if(arg_count > 1) {
		options = *args[arg_count - 1];
		convert_to_array(options);

#ifdef AST_FLAG_BLOCKING

		if(zend_hash_find(Z_ARRVAL_P(options), "play_to_caller", sizeof("play_to_caller"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			ast_set_flag(&(config.features_caller), Z_LVAL_PP(zvalue));
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "play_to_callee", sizeof("play_to_callee"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			ast_set_flag(&(config.features_caller), Z_LVAL_PP(zvalue));
		}

#else
		if(zend_hash_find(Z_ARRVAL_P(options), "play_to_caller", sizeof("play_to_caller"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			config.play_to_caller = Z_LVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "play_to_callee", sizeof("play_to_callee"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			config.play_to_callee = Z_LVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "allowredirect_in", sizeof("allowredirect_in"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			config.allowredirect_in = Z_LVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "allowredirect_out", sizeof("allowredirect_out"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			config.allowredirect_out = Z_LVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "allowdisconnect_in", sizeof("allowdisconnect_in"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			config.allowdisconnect_in = Z_LVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "allowdisconnect_out", sizeof("allowdisconnect_out"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			config.allowdisconnect_out = Z_LVAL_PP(zvalue);
		}

#endif
		if(zend_hash_find(Z_ARRVAL_P(options), "timelimit", sizeof("timelimit"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			config.timelimit = Z_LVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "play_warning", sizeof("play_warning"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			config.play_warning = Z_LVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "warning_freq", sizeof("warning_freq"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			config.warning_freq = Z_LVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "warning_sound", sizeof("warning_sound"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_string_ex(zvalue);
			config.warning_sound = Z_STRVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "end_sound", sizeof("end_sound"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_string_ex(zvalue);
			config.end_sound = Z_STRVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "start_sound", sizeof("start_sound"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_string_ex(zvalue);
			config.start_sound = Z_STRVAL_PP(zvalue);
		}

		if(zend_hash_find(Z_ARRVAL_P(options), "firstpass", sizeof("firstpass"), (void **) &zvalue) == SUCCESS)
		{
			convert_to_long_ex(zvalue);
			config.firstpass = Z_LVAL_PP(zvalue);
		}
	}

	RETURN_LONG(ast_bridge_call(chan, pr, &config));
}
/* }}} */


/* {{{ proto array asterisk_chanlist(void)
  Channel list */
PHP_FUNCTION(asterisk_chanlist)
{
	struct ast_channel *c=NULL;
	int channels = 0;
	zval *channel;

	if(ZEND_NUM_ARGS() > 0) {
		WRONG_PARAM_COUNT;
	}

	array_init(return_value);

	c = ast_channel_walk_locked(NULL);
	while(c) {
		MAKE_STD_ZVAL(channel);
		array_init(channel);

		add_assoc_string(channel, "channel",     c->name, 1);
		add_assoc_string(channel, "context",     c->context, 1);
		add_assoc_string(channel, "extension",   c->exten, 1);
		add_assoc_long(channel,   "priority",    c->priority);
		add_assoc_string(channel, "state",       ast_state2str(c->_state), 1);
		add_assoc_string(channel, "application", c->appl ? c->appl : "(None)", 1);
		add_assoc_string(channel, "data",        c->data ? (strlen(c->data) ? c->data : "(Empty)") : "(None)", 1);

		add_index_zval(return_value, channels++, channel);

		ast_mutex_unlock(&c->lock);
		c = ast_channel_walk_locked(c);
	}
}
/* }}} */


/* {{{ proto int asterisk_channel_make_compatible([string channel,] string peer)
  Make compatible */
PHP_FUNCTION(asterisk_channel_make_compatible)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();
	struct ast_channel * chan;
	struct ast_channel * pr;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_string_ex(args[arg_count - 1]);
	pr = asterisk_get_channel_by_name(Z_STRVAL_PP(args[arg_count - 1]));

	if(!pr) {
		ast_log(LOG_WARNING, "No Channels!\n");
		RETURN_LONG(-1);
	}

	RETURN_LONG(ast_channel_make_compatible(chan, pr));
}
/* }}} */


/* {{{ proto int asterisk_channel_setoption([string channel,] int option, string data)
  Set a channel option */
PHP_FUNCTION(asterisk_channel_setoption)
{
/* todo: test */
	char * channel;
	int channel_len;
	int option;
	char * data;
	int data_len;
	int arg_count = ZEND_NUM_ARGS();
	struct ast_channel * chan;

	if(arg_count < 3) {
		chan = asterisk_get_channel_by_name(NULL);
		if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ls", &option, &data, &data_len) == FAILURE) {
			return;
		}
	} else {
		if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sls", &channel, &channel_len, &option, &data, &data_len) == FAILURE) {
			return;
		}
		chan = asterisk_get_channel_by_name(channel);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	RETURN_LONG(ast_channel_setoption(chan, option, data, data_len, 0));
}
/* }}} */


/* {{{ proto int asterisk_channel_status([string channel])
  Get channel status */
PHP_FUNCTION(asterisk_channel_status)
{
	zval **args[1];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 1) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	RETURN_LONG(chan->_state);
}
/* }}} */


/* {{{ proto int asterisk_cli(int fd, string cmd)
  CLI */
PHP_FUNCTION(asterisk_cli)
{
/* TODO: test
   this command is broken
 */
	long fd = 0;
	char * cmd = NULL;
	int cmd_len = 0;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ls", &fd, &cmd, &cmd_len) == FAILURE) {
		return;
	}
/*
	ast_cli(fd, cmd);
*/
	RETURN_LONG(0);
}
/* }}} */


/* {{{ proto string asterisk_codec2str(int codec)
  CLI */
PHP_FUNCTION(asterisk_codec2str)
{
	int codec = 0;
	char * c;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &codec) == FAILURE) {
		return;
	}

	c = ast_codec2str(codec);
	RETURN_STRING(c, strlen(c));
}
/* }}} */


/* {{{ proto int asterisk_control_streamfile([string channel,] string file, string fwd, string rev, string stop, string pause[, int skipms])
  Stream a file with controls */
PHP_FUNCTION(asterisk_control_streamfile)
{
/* TODO: test */
	char * channel = NULL;
	int channel_len = 0;
	char * file = NULL;
	int file_len = 0;
	char * fwd = NULL;
	int fwd_len = 0;
	char * rev = NULL;
	int rev_len = 0;
	char * stop = NULL;
	int stop_len = 0;
	char * pause = NULL;
	int pause_len = 0;
	long skipms = 0;
	struct ast_channel * chan;

	if(ZEND_NUM_ARGS() == 5) {
		chan = asterisk_get_channel_by_name(NULL);
		if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sssss", &file, &file_len, &fwd, &fwd_len, &rev, &rev_len, &stop, &stop_len, &pause, &pause_len) == FAILURE) {
			return;
		}
	}
	else if(ZEND_NUM_ARGS() == 6) {
		chan = asterisk_get_channel_by_name(NULL);
		if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sssssl", &file, &file_len, &fwd, &fwd_len, &rev, &rev_len, &stop, &stop_len, &pause, &pause_len, &skipms) == FAILURE) {
			return;
		}
	} else {
		if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ssssssl", &channel, &channel_len, &file, &file_len, &fwd, &fwd_len, &rev, &rev_len, &stop, &stop_len, &pause, &pause_len, &skipms) == FAILURE) {
			return;
		}
		chan = asterisk_get_channel_by_name(channel);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	RETURN_LONG(ast_control_streamfile(chan, file, fwd, rev, stop, pause, skipms));
}
/* }}} */


/* {{{ proto int asterisk_db_del(string family, string keys)
  Delete a database value */
PHP_FUNCTION(asterisk_db_del)
{
/* TODO: test */
	char * family = NULL;
	int family_len = 0;
	char * keys = NULL;
	int keys_len = 0;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss", &family, &family_len, &keys, &keys_len) == FAILURE) {
		return;
	}

	RETURN_LONG(ast_db_del(family, keys));
}
/* }}} */


/* {{{ proto int asterisk_db_deltree(string family[, string keytree])
  Delete a database family */
PHP_FUNCTION(asterisk_db_deltree)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	convert_to_string_ex(args[0]);
	if(arg_count == 2) {
		convert_to_string_ex(args[1]);
		RETURN_LONG(ast_db_deltree(Z_STRVAL_PP(args[0]), Z_STRVAL_PP(args[1])));
	} else {
		RETURN_LONG(ast_db_deltree(Z_STRVAL_PP(args[0]), NULL));
	}
}
/* }}} */


/* {{{ proto int asterisk_db_get(string family, string keys)
  Get a database value */
PHP_FUNCTION(asterisk_db_get)
{
/* TODO: test
need to pass return by reference and return res
 */
	char * family = NULL;
	int family_len = 0;
	char * keys = NULL;
	int keys_len = 0;
	int res;
	char tmp[256] = {""};
	char *ptr;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss", &family, &family_len, &keys, &keys_len) == FAILURE) {
		return;
	}

	res = ast_db_get(family, keys, tmp, sizeof(tmp));
	ptr = tmp;

	if(ptr) {
		RETURN_STRING(ptr, strlen(ptr));
	}

	RETURN_NULL();
}
/* }}} */


/* {{{ proto int asterisk_db_put(string family, string keys, string value)
  Put a database value */
PHP_FUNCTION(asterisk_db_put)
{
/* TODO: test */
	char * family = NULL;
	int family_len = 0;
	char * keys = NULL;
	int keys_len = 0;
	char * value = NULL;
	int value_len = 0;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sss", &family, &family_len, &keys, &keys_len, &value, &value_len) == FAILURE) {
		return;
	}

	RETURN_LONG(ast_db_put(family, keys, value));
}
/* }}} */


/* {{{ proto int asterisk_exec([string channel,] string app[, string data])
  Execute an app */
PHP_FUNCTION(asterisk_exec)
{
	zval **args[3];
	char * data = NULL;
	char * app_name = NULL;
	int arg_count = ZEND_NUM_ARGS();
	struct ast_app *app;
	struct ast_channel * chan;

	if(arg_count > 3 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 3) {
		chan = asterisk_get_channel_by_name(NULL);
		convert_to_string_ex(args[0]);
		app_name = Z_STRVAL_PP(args[0]);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_string_ex(args[1]);
		app_name = Z_STRVAL_PP(args[1]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	app = (struct ast_app *)pbx_findapp(app_name);
	if(!app) {
		ast_log(LOG_WARNING, "Could not find application (%s)\n", app_name);
		RETURN_LONG(-1);
	}

	if(arg_count > 1) {
		convert_to_string_ex(args[arg_count - 1]);
		data = Z_STRVAL_PP(args[arg_count - 1]);
	}

	RETURN_LONG(pbx_exec(chan, app, data, 1));
}
/* }}} */


/* {{{ proto string asterisk_getdata([string channel,] string filename[, int maxdigits[, int timeout]])
  Get data */
PHP_FUNCTION(asterisk_getdata)
{
/* TODO: test
need to pass return by reference and return res
 */
	zval **args[4];
	struct ast_channel * chan;
	char * filename;
	int maxdigits = 1024;
	int timeout = 0;
	int res;
	char data[1024] = {""};
	int arg_count = ZEND_NUM_ARGS();

	if(arg_count > 4 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	convert_to_string_ex(args[0]);
	if(arg_count < 4) {
		chan = asterisk_get_channel_by_name(NULL);
		filename = Z_STRVAL_PP(args[0]);
		if(arg_count > 1) {
			convert_to_long_ex(args[1]);
			maxdigits = Z_LVAL_PP(args[1]);
			if(arg_count > 2) {
				convert_to_long_ex(args[2]);
				timeout = Z_LVAL_PP(args[2]);
			}			
		}
	} else {
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_string_ex(args[1]);
		filename = Z_STRVAL_PP(args[1]);
		convert_to_long_ex(args[2]);
		maxdigits = Z_LVAL_PP(args[2]);
		convert_to_long_ex(args[3]);
		timeout = Z_LVAL_PP(args[3]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	res = ast_app_getdata(chan, filename, data, maxdigits, timeout);
	if(!res) {
		if(data[0]) {
			RETURN_STRING(data, strlen(data));
		} else {
			RETURN_EMPTY_STRING();
		}
	}

	RETURN_LONG(res);
}
/* }}} */


/* {{{ proto int asterisk_getformatbyname(string name)
  Get format by name */
PHP_FUNCTION(asterisk_getformatbyname)
{
	char * format;
	int format_len;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &format, &format_len) == FAILURE) {
		return;
	}

	RETURN_LONG(ast_getformatbyname(format));
}
/* }}} */


/* {{{ proto string asterisk_getformatname(int format)
  Get format name */
PHP_FUNCTION(asterisk_getformatname)
{
	long format = 0;
	char * f;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &format) == FAILURE) {
		return;
	}

	f = ast_getformatname(format);
	RETURN_STRING(f, strlen(f));
}
/* }}} */


/* {{{ proto int asterisk_get_priority([string channel])
  Set priority */
PHP_FUNCTION(asterisk_get_priority)
{
	zval **args[1];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 1) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	RETURN_LONG(chan->priority);
}
/* }}} */


/* {{{ proto string asterisk_get_variable([string channel,] string variable)
  Get a channel variable */
PHP_FUNCTION(asterisk_get_variable)
{
	zval **args[2];
	char * res;
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	convert_to_string_ex(args[0]);
	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_string_ex(args[1]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	res = (char *) pbx_builtin_getvar_helper(chan, Z_STRVAL_PP(args[arg_count - 1]));

	RETURN_STRING(res, strlen(res));
}
/* }}} */


/* {{{ proto array asterisk_get_variables([string channel])
  Get channel variables */
PHP_FUNCTION(asterisk_get_variables)
{
/* TODO: test
get globals
 */
	zval **args[1];
	char * res;
	struct ast_channel * chan;
        struct ast_var_t *variables;
        struct varshead *headp;
	int arg_count = ZEND_NUM_ARGS();

	if(arg_count > 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 1) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(chan) {
		headp = &chan->varshead;
	} else {
/*
                headp = &globals;
*/
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	array_init(return_value);

	AST_LIST_TRAVERSE(headp, variables, entries) {
		add_assoc_string(return_value, variables->name, variables->value, 1);
	}
#if 0
	if(headp != &globals) {
		/* Check global variables if we haven't already */
		headp = &globals;
		AST_LIST_TRAVERSE(headp,variables,entries) {
			add_assoc_string(return_value, variables->name, variables->value, 1);
		}
	}
#endif
}
/* }}} */


/* {{{ proto int asterisk_hangup([[string channel,] int cause])
  Soft hangup a channel */
PHP_FUNCTION(asterisk_hangup)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();
	int cause = AST_SOFTHANGUP_EXPLICIT;

	struct ast_channel * chan;

	if(arg_count > 2 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(arg_count != 0) {
		convert_to_long_ex(args[arg_count - 1]);
		cause = Z_LVAL_PP(args[arg_count - 1]);
	}

	RETURN_LONG(ast_softhangup(chan, cause));
}
/* }}} */


/* {{{ proto asterisk_log(long log, string message)
  Log */
PHP_FUNCTION(asterisk_log)
{
/* TODO: test */
	int log = 0;
	char * message = NULL;
	int message_len = 0;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ls", &log, &message, &message_len) == FAILURE) {
		return;
	}

	ast_log(log, __FILE__, __LINE__, __PRETTY_FUNCTION__, message);
}
/* }}} */


/* {{{ proto int asterisk_manager_command(int fd, string cmd)
  Manager command */
PHP_FUNCTION(asterisk_manager_command)
{
/* TODO: test
optional parameters
what does this do?
broken
 */
	long fd = 0;
	char * cmd = NULL;
	int cmd_len = 0;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ls", &fd, &cmd, &cmd_len) == FAILURE) {
		return;
	}

	RETURN_LONG(ast_cli_command(fd, cmd));
}
/* }}} */


/* {{{ proto asterisk_moh_start([[string channel,] string class])
  Start music on hold */
PHP_FUNCTION(asterisk_moh_start)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_string_ex(args[arg_count - 1]);
	if(arg_count == 0 || strlen(Z_STRVAL_PP(args[arg_count - 1])) == 0) {
		ast_moh_start(chan, NULL);
	}
	else {
		ast_moh_start(chan, Z_STRVAL_PP(args[arg_count - 1]));
	}
}
/* }}} */


/* {{{ proto asterisk_moh_stop([string channel])
  Stop music on hold */
PHP_FUNCTION(asterisk_moh_stop)
{
/* TODO: test */
	zval **args[1];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 1) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	ast_moh_stop(chan);
}
/* }}} */


/* {{{ proto int asterisk_recvchar([[string channel,] int timeout])
  Receive a character */
PHP_FUNCTION(asterisk_recvchar)
{
/* TODO: test */
	zval **args[2];
	int timeout = 0;
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(arg_count) {
		convert_to_long_ex(args[arg_count - 1]);
		timeout = Z_LVAL_PP(args[arg_count - 1]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	RETURN_LONG(ast_recvchar(chan, timeout));
}
/* }}} */


/* {{{ proto string asterisk_request(string type, int format, string data, int &cause)
  Request */
PHP_FUNCTION(asterisk_request)
{
/* TODO: test */
	char * type = NULL;
	int type_len = 0;
	long format = 0;
	char * data = NULL;
	int data_len = 0;

#ifdef AST_FLAG_BLOCKING
	int cause = 0;
#endif

	struct ast_channel *chan=NULL;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sls", &type, &type_len, &format, &data, &data_len) == FAILURE) {
		return;
	}

#ifdef AST_FLAG_BLOCKING
	chan = ast_request(type, format, data, &cause);
#else
	chan = ast_request(type, format, data);
#endif

	RETURN_STRING(chan->name, strlen(chan->name));
}
/* }}} */


/* {{{ proto string asterisk_request_and_dial(string type, int format, string data, int timeoute, int outstate, string callerid)
  ??? */
PHP_FUNCTION(asterisk_request_and_dial)
{
/* TODO: test
optional parameters
needs to pass by reference
 */
	char * type = NULL;
	int type_len = 0;
	long format = 0;
	char * data = NULL;
	int data_len = 0;
	long timeout = 0;
	int outstate;
	char * callerid = NULL;

#ifdef AST_FLAG_BLOCKING
	char cidname[MAX_CALLERID_SIZE], cidnum[MAX_CALLERID_SIZE];
#endif

	int callerid_len = 0;
	struct ast_channel *chan;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "slslls", &type, &type_len, &format, &data, &data_len, &timeout, &outstate, &callerid, &callerid_len) == FAILURE) {
		return;
	}
#ifdef AST_FLAG_BLOCKING
	ast_callerid_split(callerid, cidname, sizeof(cidname), cidnum, sizeof(cidnum));
	chan = ast_request_and_dial(type, format, data, timeout, &outstate, cidnum, cidname);
#else
	chan = ast_request_and_dial(type, format, data, timeout, &outstate, callerid);
#endif

	RETURN_STRING(chan->name, strlen(chan->name));
}
/* }}} */


/* {{{ proto int asterisk_say_character_str([string channel,] string chars, [string ints[, string lang]])
  Say digits */
PHP_FUNCTION(asterisk_say_character_str)
{
/* TODO: test */
	zval **args[4];
	char * chars = NULL;
	char * ints = NULL;
	char * lang = NULL;
	int i;
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 4 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	for(i = 0; i < arg_count; i++) {
		convert_to_string_ex(args[i]);
	}

	if(arg_count < 4) {
		chan = asterisk_get_channel_by_name(NULL);
		chars = Z_STRVAL_PP(args[0]);
		if(arg_count >= 2) {
			ints = Z_STRVAL_PP(args[1]);
			if(arg_count == 3) {
				lang = Z_STRVAL_PP(args[2]);
			}
		}
	} else {
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		chars = Z_STRVAL_PP(args[1]);
		ints = Z_STRVAL_PP(args[2]);
		lang = Z_STRVAL_PP(args[3]);
		arg_count++;
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(lang == NULL) {
		lang = chan->language;
	}

	RETURN_LONG(ast_say_character_str(chan, chars, ints, lang));
}
/* }}} */


/* {{{ proto int asterisk_say_date([string channel,] [int time[, string ints[, string lang[, string format[, string timezone]]]]])
  Say the date */
PHP_FUNCTION(asterisk_say_date)
{
/* TODO: test */
	zval **args[6];
	int t;
	char * ints = NULL;
	char * lang = NULL;
	char * format = NULL;
	char * timezone = NULL;

	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 6 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 6) {
		chan = asterisk_get_channel_by_name(NULL);

		if(arg_count > 0) {
			convert_to_long_ex(args[0]);
			t = Z_LVAL_PP(args[0]);
			if(arg_count > 1) {
				convert_to_string_ex(args[1]);
				ints = Z_STRVAL_PP(args[1]);
				if(arg_count > 2) {
					convert_to_string_ex(args[2]);
					lang = Z_STRVAL_PP(args[2]);
					if(arg_count > 3) {
						convert_to_string_ex(args[3]);
						format = Z_STRVAL_PP(args[3]);
						if(arg_count > 4) {
							convert_to_string_ex(args[4]);
							timezone = Z_STRVAL_PP(args[4]);
						}
					}
				}
			}
		} else {
			t = time(NULL);
		}
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_long_ex(args[1]);
		t = Z_LVAL_PP(args[1]);
		convert_to_string_ex(args[2]);
		ints = Z_STRVAL_PP(args[2]);
		convert_to_string_ex(args[3]);
		lang = Z_STRVAL_PP(args[3]);
		convert_to_string_ex(args[4]);
		format = Z_STRVAL_PP(args[4]);
		convert_to_string_ex(args[5]);
		timezone = Z_STRVAL_PP(args[5]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(lang == NULL) {
		lang = chan->language;
	}

	if(arg_count > 3) {
		RETURN_LONG(ast_say_date_with_format(chan, t, ints, lang, format, timezone));
	} else {
		RETURN_LONG(ast_say_date(chan, t, ints, lang));
	}
}
/* }}} */


/* {{{ proto int asterisk_say_datetime([string channel,] [int time[, string ints[, string lang]]])
  Say the date and time */
PHP_FUNCTION(asterisk_say_datetime)
{
/* TODO: test */
	zval **args[4];
	int t;
	char * ints = NULL;
	char * lang = NULL;

	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 4 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 4) {
		chan = asterisk_get_channel_by_name(NULL);

		if(arg_count > 0) {
			convert_to_long_ex(args[0]);
			t = Z_LVAL_PP(args[0]);
			if(arg_count > 1) {
				convert_to_string_ex(args[1]);
				ints = Z_STRVAL_PP(args[1]);
				if(arg_count > 2) {
					convert_to_string_ex(args[2]);
					lang = Z_STRVAL_PP(args[2]);
				}
			}
		} else {
			t = time(NULL);
		}
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_long_ex(args[1]);
		t = Z_LVAL_PP(args[1]);
		convert_to_string_ex(args[2]);
		ints = Z_STRVAL_PP(args[2]);
		convert_to_string_ex(args[3]);
		lang = Z_STRVAL_PP(args[3]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(lang == NULL) {
		lang = chan->language;
	}

	RETURN_LONG(ast_say_datetime(chan, t, ints, lang));
}
/* }}} */


/* {{{ proto int asterisk_say_datetime_from_now([string channel,] [int time[, string ints[, string lang]]])
  Say the date and time from now*/
PHP_FUNCTION(asterisk_say_datetime_from_now)
{
/* TODO: test */
	zval **args[4];
	int t;
	char * ints = NULL;
	char * lang = NULL;

	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 4 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 4) {
		chan = asterisk_get_channel_by_name(NULL);

		if(arg_count > 0) {
			convert_to_long_ex(args[0]);
			t = Z_LVAL_PP(args[0]);
			if(arg_count > 1) {
				convert_to_string_ex(args[1]);
				ints = Z_STRVAL_PP(args[1]);
				if(arg_count > 2) {
					convert_to_string_ex(args[2]);
					lang = Z_STRVAL_PP(args[2]);
				}
			}
		} else {
			t = time(NULL);
		}
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_long_ex(args[1]);
		t = Z_LVAL_PP(args[1]);
		convert_to_string_ex(args[2]);
		ints = Z_STRVAL_PP(args[2]);
		convert_to_string_ex(args[3]);
		lang = Z_STRVAL_PP(args[3]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(lang == NULL) {
		lang = chan->language;
	}

	RETURN_LONG(ast_say_datetime_from_now(chan, t, ints, lang));
}
/* }}} */


/* {{{ proto int asterisk_say_digit_str([string channel,] string digits, [string ints[, string lang="en"]])
  Say digits */
PHP_FUNCTION(asterisk_say_digit_str)
{
/* TODO: test */
	zval **args[4];
	char * digits = NULL;
	char * ints = NULL;
	char * lang = NULL;
	int i;
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 4 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	for(i = 0; i < arg_count; i++) {
		convert_to_string_ex(args[i]);
	}

	if(arg_count < 4) {
		chan = asterisk_get_channel_by_name(NULL);
		digits = Z_STRVAL_PP(args[0]);
		if(arg_count >= 2) {
			ints = Z_STRVAL_PP(args[1]);
			if(arg_count == 3) {
				lang = Z_STRVAL_PP(args[2]);
			}
		}
	} else {
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		digits = Z_STRVAL_PP(args[1]);
		ints = Z_STRVAL_PP(args[2]);
		lang = Z_STRVAL_PP(args[3]);
		arg_count++;
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(lang == NULL) {
		lang = chan->language;
	}

	RETURN_LONG(ast_say_digit_str(chan, digits, ints, lang));
}
/* }}} */


/* {{{ proto int asterisk_say_number([string channel], int number, [string ints[, string lang[, string options]]])
  Say a number */
PHP_FUNCTION(asterisk_say_number)
{
/* TODO: test */
	zval **args[5];
	int number;
	char * ints = NULL;
	char * lang = NULL;
	char * options = NULL;

	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 5 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 5) {
		chan = asterisk_get_channel_by_name(NULL);
		convert_to_long_ex(args[0]);
		number = Z_LVAL_PP(args[0]);
		if(arg_count >= 2) {
			convert_to_string_ex(args[1]);
			ints = Z_STRVAL_PP(args[1]);
			if(arg_count >= 3) {
				convert_to_string_ex(args[2]);
				lang = Z_STRVAL_PP(args[2]);
				if(arg_count >= 4) {
					convert_to_string_ex(args[3]);
					options = Z_STRVAL_PP(args[3]);
				}
			}
		}
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_long_ex(args[1]);
		number = Z_LVAL_PP(args[1]);
		convert_to_string_ex(args[2]);
		ints = Z_STRVAL_PP(args[2]);
		convert_to_string_ex(args[3]);
		lang = Z_STRVAL_PP(args[3]);
		convert_to_string_ex(args[4]);
		options = Z_STRVAL_PP(args[4]);
		arg_count++;
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(lang == NULL) {
		lang = chan->language;
	}

	RETURN_LONG(ast_say_number(chan, number, ints, lang, options));
}
/* }}} */


/* {{{ proto int asterisk_say_phonetic_str([string channel,] string text[, string ints[, string lang]])
  Say phonetic */
PHP_FUNCTION(asterisk_say_phonetic_str)
{
/* TODO: test */
	zval **args[4];
	char * text;
	char * ints = NULL;
	char * lang = NULL;

	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 4 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 4) {
		chan = asterisk_get_channel_by_name(NULL);
		convert_to_string_ex(args[0]);
		text = Z_STRVAL_PP(args[0]);
		if(arg_count > 1) {
			convert_to_string_ex(args[1]);
			ints = Z_STRVAL_PP(args[1]);
			if(arg_count > 2) {
				convert_to_string_ex(args[2]);
				lang = Z_STRVAL_PP(args[2]);
			}
		}
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_string_ex(args[1]);
		text = Z_STRVAL_PP(args[1]);
		convert_to_string_ex(args[2]);
		ints = Z_STRVAL_PP(args[2]);
		convert_to_string_ex(args[3]);
		lang = Z_STRVAL_PP(args[3]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(lang == NULL) {
		lang = chan->language;
	}

	RETURN_LONG(ast_say_phonetic_str(chan, text, ints, lang));
}
/* }}} */


/* {{{ proto int asterisk_say_time([string channel,] [int time[, string ints[, string lang]]])
  Say the time */
PHP_FUNCTION(asterisk_say_time)
{
/* TODO: test */
	zval **args[4];
	int t;
	char * ints = NULL;
	char * lang = NULL;

	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 4 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 4) {
		chan = asterisk_get_channel_by_name(NULL);

		if(arg_count > 0) {
			convert_to_long_ex(args[0]);
			t = Z_LVAL_PP(args[0]);
			if(arg_count > 1) {
				convert_to_string_ex(args[1]);
				ints = Z_STRVAL_PP(args[1]);
				if(arg_count > 2) {
					convert_to_string_ex(args[2]);
					lang = Z_STRVAL_PP(args[2]);
				}
			}
		} else {
			t = time(NULL);
		}
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_long_ex(args[1]);
		t = Z_LVAL_PP(args[1]);
		convert_to_string_ex(args[2]);
		ints = Z_STRVAL_PP(args[2]);
		convert_to_string_ex(args[3]);
		lang = Z_STRVAL_PP(args[3]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(lang == NULL) {
		lang = chan->language;
	}

	RETURN_LONG(ast_say_time(chan, t, ints, lang));
}
/* }}} */


/* {{{ proto int asterisk_senddigit([string channel,] string digit)
  Send a dtmf digit */
PHP_FUNCTION(asterisk_senddigit)
{
/* TODO: test */
	zval **args[2];
	char * digit;
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_string_ex(args[arg_count - 1]);
	digit = Z_STRVAL_PP(args[arg_count - 1]);

	RETURN_LONG(ast_senddigit(chan, digit[0]));
}
/* }}} */


/* {{{ proto int asterisk_send_image([string channel,] string image)
  Send an image */
PHP_FUNCTION(asterisk_send_image)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_string_ex(args[arg_count - 1]);
	RETURN_LONG(ast_send_image(chan, Z_STRVAL_PP(args[arg_count - 1])));
}
/* }}} */


/* {{{ proto int asterisk_sendtext([string channel,] string text)
  Send text */
PHP_FUNCTION(asterisk_sendtext)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_string_ex(args[arg_count - 1]);
	RETURN_LONG(ast_sendtext(chan, Z_STRVAL_PP(args[arg_count - 1])));
}
/* }}} */


/* {{{ proto asterisk_set_best_read_format([string channel])
  Set best read format */
PHP_FUNCTION(asterisk_set_best_read_format)
{
/* TODO: test
needs a parameter, renamed?
 */
	zval **args[1];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count) {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	} else {
		chan = asterisk_get_channel_by_name(NULL);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	ast_set_read_format_a(chan, ast_best_codec(chan->nativeformats));
}
/* }}} */


/* {{{ proto asterisk_set_best_write_format([string channel])
  Set best write format */
PHP_FUNCTION(asterisk_set_best_write_format)
{
/* TODO: test
needs a parameter, renamed?
 */
	zval **args[1];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count) {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	} else {
		chan = asterisk_get_channel_by_name(NULL);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	ast_set_write_format_a(chan, ast_best_codec(chan->nativeformats));
}
/* }}} */


/* {{{ proto asterisk_set_callerid([string channel,] string callerid[, boolean anitoo])
  Set Caller ID */
PHP_FUNCTION(asterisk_set_callerid)
{
/* TODO: test */
	zval **args[3];
	char * callerid = NULL;
	int anitoo = 0;
	int arg_count = ZEND_NUM_ARGS();

#ifdef AST_FLAG_BLOCKING
	char ourcidname[MAX_CALLERID_SIZE], ourcidnum[MAX_CALLERID_SIZE];
#endif

	struct ast_channel * chan;

	if(arg_count > 3 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	convert_to_string_ex(args[0]);
	if(arg_count < 3) {
		chan = asterisk_get_channel_by_name(NULL);
		callerid = Z_STRVAL_PP(args[0]);
		if(arg_count > 1) {
			convert_to_boolean_ex(args[1]);
			anitoo = Z_BVAL_PP(args[1]);
		}
	} else {
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_string_ex(args[1]);
		callerid = Z_STRVAL_PP(args[1]);
		convert_to_boolean_ex(args[2]);
		anitoo = Z_BVAL_PP(args[2]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

#ifdef AST_FLAG_BLOCKING
	ast_callerid_split(callerid, ourcidname, sizeof(ourcidname), ourcidnum, sizeof(ourcidnum));
	ast_set_callerid(chan, ourcidnum, ourcidname, ourcidnum);
#else
	ast_set_callerid(chan, callerid, anitoo);
#endif

}
/* }}} */


/* {{{ proto int asterisk_set_context([string channel,] string context)
  Set context */
PHP_FUNCTION(asterisk_set_context)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_string_ex(args[arg_count - 1]);
	strncpy(chan->context, Z_STRVAL_PP(args[arg_count - 1]), sizeof(chan->context)-1);
	RETURN_LONG(1);
}
/* }}} */


/* {{{ proto int asterisk_set_extension([string channel,] string exten)
  Set extension */
PHP_FUNCTION(asterisk_set_extension)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_string_ex(args[arg_count - 1]);
	strncpy(chan->exten, Z_STRVAL_PP(args[arg_count - 1]), sizeof(chan->exten)-1);
	RETURN_LONG(0);
}
/* }}} */


/* {{{ proto int asterisk_set_priority([string channel,] int priority)
  Set priority */
PHP_FUNCTION(asterisk_set_priority)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_long_ex(args[arg_count - 1]);
	chan->priority = Z_LVAL_PP(args[arg_count - 1]) - 1;
	RETURN_LONG(1);
}
/* }}} */


/* {{{ proto asterisk_set_read_format([string channel,] int format)
  Set read format */
PHP_FUNCTION(asterisk_set_read_format)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_long_ex(args[arg_count - 1]);
	ast_set_read_format_a(chan, Z_LVAL_PP(args[arg_count - 1]));
}
/* }}} */


/* {{{ proto int asterisk_set_variable([string channel,] string variable, string value)
  Set a channel variable */
PHP_FUNCTION(asterisk_set_variable)
{
	zval **args[3];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 3 || arg_count < 2 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 3) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_string_ex(args[arg_count - 2]);
	convert_to_string_ex(args[arg_count - 1]);
	pbx_builtin_setvar_helper(chan, Z_STRVAL_PP(args[arg_count - 2]), Z_STRVAL_PP(args[arg_count - 1]));
	RETURN_LONG(0);
}
/* }}} */


/* {{{ proto int asterisk_setwhentohangup([[string channel,] int timeout])
  Cause the channel to automatically hangup at <time> seconds in the future. */
PHP_FUNCTION(asterisk_setwhentohangup)
{
/* todo: use ast_channel_setwhentohangup instead? */
	zval **args[2];
	time_t timeout = 0;
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(arg_count > 0)
	{
		convert_to_long_ex(args[arg_count - 1]);
		timeout = Z_LVAL_PP(args[arg_count - 1]);
		if(timeout < 0) {
			timeout = 0;
		}
	}

	ast_channel_setwhentohangup(chan, timeout);

	RETURN_LONG(0);
}
/* }}} */


/* {{{ proto asterisk_set_write_format([string channel,] int format)
  Set write format */
PHP_FUNCTION(asterisk_set_write_format)
{
/* TODO: test */
	zval **args[2];
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_string_ex(args[arg_count - 1]);
	ast_set_write_format_a(chan, Z_LVAL_PP(args[arg_count - 1]));
}
/* }}} */


/* {{{ proto string asterisk_state2str(int state)
  CLI */
PHP_FUNCTION(asterisk_state2str)
{
	int state = 0;
	char * s;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &state) == FAILURE) {
		return;
	}

	s = ast_state2str(state);
	RETURN_STRING(s, strlen(s));
}
/* }}} */


/* {{{ proto int asterisk_stream_file([string channel,] string filename[, string digits[, int offset[, string lang]]])
  Stream a file */
PHP_FUNCTION(asterisk_stream_file)
{
/* TODO: test
broke!
pass offset back by reference?
 */
	zval **args[5];
	char * filename = NULL;
	char * digits = NULL;
	long offset = 0;
	char * lang = NULL;
	int res = 0;
	struct ast_filestream *fs;
	long max_length;
	int input;
	struct ast_channel * chan;
	int arg_count = ZEND_NUM_ARGS();

	if(arg_count > 5 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	convert_to_string_ex(args[0]);
	if(arg_count < 5) {
		chan = asterisk_get_channel_by_name(NULL);
		filename = Z_STRVAL_PP(args[0]);
		if(arg_count > 1) {
			convert_to_string_ex(args[1]);
			digits = Z_STRVAL_PP(args[1]);
			if(arg_count > 2) {
				convert_to_long_ex(args[2]);
				offset = Z_LVAL_PP(args[2]);
				if(arg_count > 3) {
					convert_to_string_ex(args[3]);
					lang = Z_STRVAL_PP(args[3]);
				}
			}
		}
	} else {
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
		convert_to_string_ex(args[1]);
		filename = Z_STRVAL_PP(args[1]);
		convert_to_string_ex(args[2]);
		digits = Z_STRVAL_PP(args[2]);
		convert_to_long_ex(args[3]);
		offset = Z_LVAL_PP(args[3]);
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(!lang) {
		lang = chan->language;
	}

	if(!digits) {
		digits = "";
	}

	res = ast_waitfor(chan, -1);
	input = ast_waitfordigit(chan, 100);
	if(input < 0) {
		RETURN_LONG(-1);
	}

	fs = (struct ast_filestream *) ast_openstream(chan, filename, lang);
	if(!fs) {
		ast_log(LOG_WARNING, "Unable to open %s\n", filename);
		RETURN_LONG(-1);
	}
	ast_seekstream(fs, 0, SEEK_END);
	max_length = ast_tellstream(fs);
	ast_seekstream(fs, offset, SEEK_SET);
	res = ast_applystream(chan, fs);
	res = ast_playstream(fs);
	if(res) {
		RETURN_LONG(-1);
	}
	res = ast_waitstream(chan, digits);
	offset = (chan->stream) ? ast_tellstream(fs) : max_length;
	ast_stopstream(chan);

	RETURN_LONG(res);
}
/* }}} */


/* {{{ proto int asterisk_tdd_mode([[string channel,] int mode])
  Set tdd mode */
PHP_FUNCTION(asterisk_tdd_mode)
{
/* TODO: test */
	zval **args[2];
	int mode = 1;
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(arg_count) {
		convert_to_long_ex(args[arg_count - 1]);
		mode = Z_LVAL_PP(args[arg_count - 1]);
	}

	RETURN_LONG(ast_channel_setoption(chan, AST_OPTION_TDD, &mode, sizeof(char), 0));
}
/* }}} */


/* {{{ proto asterisk_verbose(int level, string message)
  Verbose a message to the CLI */
PHP_FUNCTION(asterisk_verbose)
{
	int level = 0;
	char * message = NULL;
	int message_len = 0;
	char *prefix[] = {"", VERBOSE_PREFIX_1, VERBOSE_PREFIX_2, VERBOSE_PREFIX_3, VERBOSE_PREFIX_4};
	struct ast_channel * chan;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ls", &level, &message, &message_len) == FAILURE) {
		return;
	}

	if(level >= 0 && level < 4 && level <= option_verbose) {
		chan = asterisk_get_channel_by_name(NULL);

		if(chan) {
			ast_verbose("%s %s: %s\n", prefix[level], chan->data, message);
		} else {
			ast_verbose("%s PHP: %s\n", prefix[level], message);
		}
	}
}
/* }}} */


/* {{{ proto asterisk_waitfor([string channel,] int wait)
  Wait */
PHP_FUNCTION(asterisk_waitfor)
{
/* TODO: test
what is wait?
 */
	zval **args[2];
	char * mode;
	int m = 0;
	int arg_count = ZEND_NUM_ARGS();

	struct ast_channel * chan;

	if(arg_count > 2 || arg_count < 1 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	convert_to_long_ex(args[arg_count - 1]);
	ast_waitfor(chan, Z_LVAL_PP(args[arg_count - 1]));
}
/* }}} */


/* {{{ proto int asterisk_waitfordigit([[string channel,] int timeout])
  Wait for a digit */
PHP_FUNCTION(asterisk_waitfordigit)
{
/* TODO: test */
	zval **args[2];
	int timeout = -1;
	int arg_count = ZEND_NUM_ARGS();
	struct ast_channel * chan;

	if(arg_count > 2 || zend_get_parameters_array_ex(arg_count, args) == FAILURE) {
		WRONG_PARAM_COUNT;
	}

	if(arg_count < 2) {
		chan = asterisk_get_channel_by_name(NULL);
	} else {
		convert_to_string_ex(args[0]);
		chan = asterisk_get_channel_by_name(Z_STRVAL_PP(args[0]));
	}

	if(!chan) {
		ast_log(LOG_WARNING, "No Channel!\n");
		RETURN_LONG(-1);
	}

	if(arg_count) {
		convert_to_long_ex(args[arg_count - 1]);
		timeout = Z_LVAL_PP(args[arg_count - 1]);
	}

	RETURN_LONG(ast_waitfordigit(chan, timeout));
}
/* }}} */


/* *********************************************************************** *
 * *            PHP SAPI                                                 * *
 * *********************************************************************** */



/*
 * sapi_asterisk_ub_write: Write len bytes to the connection output.
 */
static int sapi_asterisk_ub_write(const char *str, unsigned int str_length TSRMLS_DC)
{

  FILE *fp = fopen("/tmp/asterisk_php.log", "a");
  fwrite(str, str_length, sizeof(char), fp);
  fclose(fp);


/* todo:  fix me! */
	char buffer[4096];
	int i, j = 0;
	for(i = 0; i < str_length; i++) {
		buffer[j++] = str[i];
		if(str[i] == 10) { /* new line */
			buffer[j] = 0;
			ast_verbose(VERBOSE_PREFIX_2 "%s", buffer);
			j = 0;
		}
		else if(str[i] == 0) { /* null character */
			ast_verbose("%s", buffer);
			j = 0;
		}
		if(j == 4095) { /* don't overfill buffer */
			buffer[j] = 0;
			ast_verbose("%s", buffer);
			j = 0;
		}
	}
	if(j) { /* stuff left over */
		buffer[j] = 0;
		ast_verbose("%s", buffer);
	}
	return str_length;
}

void asterisk_error_handler(int type, const char *error_filename, const uint error_lineno, const char *format, va_list args)
{
	char *buffer;
	int buffer_len;
	TSRMLS_FETCH();

	buffer_len = vspprintf(&buffer, PG(log_errors_max_len), format, args);

	/* display/log the error if necessary */
	if((EG(error_reporting) & type || (type & E_CORE)) && (PG(log_errors) || PG(display_errors))) {
		char *error_type_str;

		switch (type) {
			case E_ERROR:
			case E_CORE_ERROR:
			case E_COMPILE_ERROR:
			case E_USER_ERROR:
				error_type_str = "Fatal error";
				break;
			case E_WARNING:
			case E_CORE_WARNING:
			case E_COMPILE_WARNING:
			case E_USER_WARNING:
				error_type_str = "Warning";
				break;
			case E_PARSE:
				error_type_str = "Parse error";
				break;
			case E_NOTICE:
			case E_USER_NOTICE:
				error_type_str = "Notice";
				break;
			default:
				error_type_str = "Unknown error";
				break;
		}

		if(PG(log_errors)) {
			char *log_buffer;
#ifdef PHP_WIN32
			if(type == E_CORE_ERROR || type == E_CORE_WARNING) {
				MessageBox(NULL, buffer, error_type_str, MB_OK|ZEND_SERVICE_MB_STYLE);
			}
#endif
			spprintf(&log_buffer, 0, "PHP %s:  %s in %s on line %d", error_type_str, buffer, error_filename, error_lineno);
			php_log_err(log_buffer TSRMLS_CC);
			efree(log_buffer);
		}
 
		if(PG(display_errors)) {
			char *prepend_string = INI_STR("error_prepend_string");
			char *append_string = INI_STR("error_append_string");
			char *error_format = "%s\n%s: %s in %s on line %d\n%s";    
			ast_verbose(error_format, STR_PRINT(prepend_string), error_type_str, buffer, error_filename, error_lineno, STR_PRINT(append_string));
		}
	}

	/* Bail out if we can't recover */
	switch(type) {
		case E_CORE_ERROR:
		case E_ERROR:
		/*case E_PARSE: the parser would return 1 (failure), we can bail out nicely */
		case E_COMPILE_ERROR:
		case E_USER_ERROR:
			ast_verbose("\nPHP: %s exiting\n", error_filename);
			EG(exit_status) = 255;
#if MEMORY_LIMIT
			/* restore memory limit */
			AG(memory_limit) = PG(memory_limit); 
#endif
			efree(buffer);
			zend_bailout();
			return;
			break;
	}

	/* Log if necessary */
	if(PG(track_errors) && EG(active_symbol_table)) {
		pval *tmp;

		ALLOC_ZVAL(tmp);
		INIT_PZVAL(tmp);
		Z_STRVAL_P(tmp) = (char *) estrndup(buffer, buffer_len);
		Z_STRLEN_P(tmp) = buffer_len;
		Z_TYPE_P(tmp) = IS_STRING;
		zend_hash_update(EG(active_symbol_table), "php_errormsg", sizeof("php_errormsg"), (void **) & tmp, sizeof(pval *), NULL);
	}
	efree(buffer);
}


/*
 * sapi_asterisk_header_handler: Add/update response headers with those provided by the PHP engine.
 */
static int sapi_asterisk_header_handler(sapi_header_struct * sapi_header, sapi_headers_struct * sapi_headers TSRMLS_DC)
{
	return 0;
}

/*
 * sapi_asterisk_send_headers: Transmit the headers to the client. This has the
 * effect of starting the response under asterisk.
 */
static int sapi_asterisk_send_headers(sapi_headers_struct * sapi_headers TSRMLS_DC)
{
	return SAPI_HEADER_SENT_SUCCESSFULLY;
}

static int sapi_asterisk_read_post(char *buffer, uint count_bytes TSRMLS_DC)
{
	return 0;
}

/*
 * sapi_asterisk_read_cookies: Return cookie information into PHP.
 */
static char *sapi_asterisk_read_cookies(TSRMLS_D)
{
	return NULL;
}

static void sapi_asterisk_register_server_variables(zval * track_vars_array TSRMLS_DC)
{
#if 0
   php_asterisk_globals *rc = (php_asterisk_globals *) SG(server_context);
   size_t i;
   char *value;
   char buf[128];

   /* PHP_SELF and REQUEST_URI */
   value = lstFset_get(rc->t->vars, "uri");
   if(value != NULL) {
      php_register_variable("PHP_SELF", value, track_vars_array TSRMLS_CC);
      php_register_variable("REQUEST_URI", value, track_vars_array TSRMLS_CC);
   }
 
   /* COUNTRY CODE */
   value = lstFset_get(rc->t->vars, "ccode");
   if(value!=NULL)
     php_register_variable("COUNTRY_CODE", value, track_vars_array TSRMLS_CC);

   /* argv */
   value = lstFset_get(rc->t->vars, "query");
   if(value != NULL)
      php_register_variable("argv", value, track_vars_array TSRMLS_CC);

   /* GATEWAY_INTERFACE */
   php_register_variable("GATEWAY_INTERFACE", "CGI/1.1", track_vars_array TSRMLS_CC);

   /* SERVER_NAME and HTTP_HOST */
   value = lstFset_get(rc->t->req_hdrs, "host");
   if(value != NULL) {
      php_register_variable("HTTP_HOST", value, track_vars_array TSRMLS_CC);
      /* TODO: This should probably scrub the port value if one is present. */
      php_register_variable("SERVER_NAME", value, track_vars_array TSRMLS_CC);
   }
   /* SERVER_SOFTWARE */
   value = lstFset_get(rc->t->res_hdrs, "Server");
   if(value != NULL)
      php_register_variable("SERVER_SOFTWARE", value, track_vars_array TSRMLS_CC);

   /* SERVER_PROTOCOL */
   value = lstFset_get(rc->t->vars, "protocol");
   if(value != NULL)
      php_register_variable("SERVER_PROTOCOL", value, track_vars_array TSRMLS_CC);

   /* REQUEST_METHOD */
   value = lstFset_get(rc->t->vars, "method");
   if(value != NULL)
      php_register_variable("REQUEST_METHOD", value, track_vars_array TSRMLS_CC);

   /* QUERY_STRING */
   value = lstFset_get(rc->t->vars, "query");
   if(value != NULL)
      php_register_variable("QUERY_STRING", value, track_vars_array TSRMLS_CC);

   /* DOCUMENT_ROOT */
   value = lstFset_get(rc->t->vars, "docroot");
   if(value != NULL)
      php_register_variable("DOCUMENT_ROOT", value, track_vars_array TSRMLS_CC);

   /* HTTP_ACCEPT */
   value = lstFset_get(rc->t->req_hdrs, "accept");
   if(value != NULL)
      php_register_variable("HTTP_ACCEPT", value, track_vars_array TSRMLS_CC);

   /* HTTP_ACCEPT_CHARSET */
   value = lstFset_get(rc->t->req_hdrs, "accept-charset");
   if(value != NULL)
      php_register_variable("HTTP_ACCEPT_CHARSET", value, track_vars_array TSRMLS_CC);

   /* HTTP_ACCEPT_ENCODING */
   value = lstFset_get(rc->t->req_hdrs, "accept-encoding");
   if(value != NULL)
      php_register_variable("HTTP_ACCEPT_ENCODING", value, track_vars_array TSRMLS_CC);

   /* HTTP_ACCEPT_LANGUAGE */
   value = lstFset_get(rc->t->req_hdrs, "accept-language");
   if(value != NULL)
      php_register_variable("HTTP_ACCEPT_LANGUAGE", value, track_vars_array TSRMLS_CC);

   /* HTTP_CONNECTION */
   value = lstFset_get(rc->t->req_hdrs, "connection");
   if(value != NULL)
      php_register_variable("HTTP_CONNECTION", value, track_vars_array TSRMLS_CC);

   /* HTTP_REFERER */
   value = lstFset_get(rc->t->req_hdrs, "referer");
   if(value != NULL)
      php_register_variable("HTTP_REFERER", value, track_vars_array TSRMLS_CC);

   /* HTTP_USER_AGENT */
   value = lstFset_get(rc->t->req_hdrs, "user-agent");
   if(value != NULL)
      php_register_variable("HTTP_USER_AGENT", value, track_vars_array TSRMLS_CC);

   /* REMOTE_ADDR */
   utlFip_to_str(rc->t->cli_ipv4_addr, buf, sizeof(buf));
   php_register_variable("REMOTE_ADDR", buf, track_vars_array TSRMLS_CC);

   /* REMOTE_PORT */

   /* SCRIPT_FILENAME and PATH_TRANSLATED */
   value = lstFset_get(rc->t->vars, "path");
   if(value != NULL) {
      php_register_variable("SCRIPT_FILENAME", value, track_vars_array TSRMLS_CC);
      php_register_variable("PATH_TRANSLATED", value, track_vars_array TSRMLS_CC);
   }
   /* SERVER_ADMIN */
   /* Not applicable */

   /* SERVER_PORT */
#endif
}

static void asterisk_log_message(char *message)
{
	ast_verbose("%s\n", message);
}

static int php_asterisk_startup(sapi_module_struct *sapi_module);

sapi_module_struct asterisk_sapi_module = {
   "asterisk",					/* name */
   "Asterisk",					/* pretty name */

   php_asterisk_startup,			/* startup */
   php_module_shutdown_wrapper,			/* shutdown */

   NULL,					/* activate */
   NULL,					/* deactivate */

   sapi_asterisk_ub_write,			/* unbuffered write */
   NULL,					/* flush */
   NULL,					/* get uid */
   NULL,					/* getenv */

   php_error,					/* error handler */

   sapi_asterisk_header_handler,		/* header handler */
   sapi_asterisk_send_headers,			/* send headers handler */
   NULL,					/* send header handler */

   sapi_asterisk_read_post,			/* read POST data */
   sapi_asterisk_read_cookies,			/* read Cookies */

   sapi_asterisk_register_server_variables,	/* register server variables */
   asterisk_log_message,			/* Log message */
   NULL,					/* Get request time */

   NULL,					/* Block interruptions */
   NULL,					/* Unblock interruptions */

   STANDARD_SAPI_MODULE_PROPERTIES
};

static int php_asterisk_startup(sapi_module_struct *sapi_module)
{
	if(php_module_startup(sapi_module, &asterisk_module_entry, 1) == FAILURE) {
		return FAILURE;
	}
	return SUCCESS;
}

/* *********************************************************************** *
 * *            Asterisk Application                                     * *
 * *********************************************************************** */


static char *tdesc      = "PHP for the dialplan";
static char *app        = "PHP";
static char *synopsis   = "Execute PHP code";
static char *descrip    = "  PHP(filename[|option]):  Executes filename as PHP code.\n";

/*
int ast_register_string_constant(char *var, char *val)
{
	TSRMLS_FETCH();
	zend_constant c;

	c.value.type = IS_STRING;
	c.value.value.str.val = zend_strndup(val, strlen(val));
	c.value.value.str.len = strlen(val);
	c.flags = CONST_CS;
	c.name = zend_strndup(var, strlen(var));
	c.name_len = strlen(var);
	c.module_number = PHP_USER_CONSTANT;
	if(zend_register_constant(&c TSRMLS_CC) == SUCCESS) {
		return 1;
	} else {
		return 0;
	}
}
*/

static void asterisk_request_ctor(struct php_asterisk_globals *request_context TSRMLS_DC)
{
/*	ast_register_string_constant("JELLO", "FISH"); */

	zend_error_cb = asterisk_error_handler;

	SG(request_info).argc = request_context->argc;
	SG(request_info).argv = request_context->argv;

	SG(request_info).path_translated    = estrdup(request_context->argv[0]);

#if 0
   SG(request_info).query_string       = estrdup();
   SG(request_info).request_uri        = estrdup(uri);
   SG(request_info).request_method     = estrdup(request_method);
   SG(request_info).content_type       = estrdup(content_type);
   SG(request_info).content_length     = (content_length == NULL) ? 0 : strtoul(content_length, 0, 0);
   SG(sapi_headers).http_response_code = 200;
#endif
}

static void asterisk_request_dtor(struct php_asterisk_globals *request_context TSRMLS_DC)
{
	efree(SG(request_info).path_translated);
#if 0
   efree(SG(request_info).query_string);
   efree(SG(request_info).request_uri);
   efree(SG(request_info).request_method);
   efree(SG(request_info).content_type);
#endif
}

int asterisk_module_main(struct php_asterisk_globals *request_context TSRMLS_DC)
{
	int retval;
	zend_file_handle file_handle;

	if(php_request_startup(TSRMLS_C) == FAILURE) {
		return FAILURE;
	}

	file_handle.type = ZEND_HANDLE_FILENAME;
	file_handle.filename = SG(request_info).path_translated;
	file_handle.free_filename = 0;
	file_handle.opened_path = NULL;

	retval = php_execute_script(&file_handle TSRMLS_CC);
	php_request_shutdown(NULL);

	return retval;
}

static int asterisk_php_exec(struct ast_channel *chan, void *data)
{
	int retval = 0;
	struct localuser *u;
	char buf[2048] = "";
	char *tmp = (char *)buf;
	char *stringp;
	php_asterisk_globals *request_context;

	TSRMLS_FETCH();

	request_context = (php_asterisk_globals *) malloc(sizeof(php_asterisk_globals));

	request_context->channel = chan;

	request_context->argc = 0;

	if(!data || ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "PHP requires an argument (script)\n");
		return -1;
	}
	strncpy(buf, data, sizeof(buf) - 1);

	while((request_context->argc < MAX_ARGS && (stringp = strsep(&tmp, "|")))) {
		request_context->argv[request_context->argc++] = stringp;
	}
	request_context->argv[request_context->argc] = NULL;

	LOCAL_USER_ADD(u);

	SG(server_context) = request_context;

	asterisk_request_ctor(request_context TSRMLS_CC);
	retval = asterisk_module_main(request_context TSRMLS_CC);
	asterisk_request_dtor(request_context TSRMLS_CC);

	free(request_context);

	/*
	 * This call is ostensibly provided to free the memory from PHP/TSRM when
	 * the thread terminated, but, it leaks a structure in some hash list
	 * according to the developers. Not calling this will leak the entire
	 * interpreter, around 100k, but calling it and then terminating the
	 * thread will leak the struct (around a k). The only answer with the
	 * current TSRM implementation is to reuse the threads that allocate TSRM
	 * resources.
	 */
	/* ts_free_thread(); */

	return retval;
}

int load_module(void)
{
	php_core_globals *core_globals;

	tsrm_startup(128, 1, 0, NULL);
	core_globals = ts_resource(core_globals_id);

	sapi_startup(&asterisk_sapi_module);
	asterisk_sapi_module.startup(&asterisk_sapi_module);

	return ast_register_application(app, asterisk_php_exec, synopsis, descrip);
}

int unload_module(void)
{
/*
	if(SG(server_context) != NULL) {
		asterisk_sapi_module.shutdown(&asterisk_sapi_module);
		sapi_shutdown();
	}
*/
	STANDARD_HANGUP_LOCALUSERS;
	return ast_unregister_application(app);
}


char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}

